<?php

/*
 * This file is part of the Fxp Composer Asset Plugin package.
 *
 * (c) François Pluchino <francois.pluchino@gmail.com>
 *
 * For the full copyright and license information, please view the LICENSE
 * file that was distributed with this source code.
 */

namespace Fxp\Composer\AssetPlugin\Repository;

use Composer\Config;
use Composer\DependencyResolver\Pool;
use Composer\EventDispatcher\EventDispatcher;
use Composer\IO\IOInterface;
use Composer\Package\CompletePackageInterface;
use Composer\Package\Loader\ArrayLoader;
use Composer\Repository\ArrayRepository;
use Fxp\Composer\AssetPlugin\Exception\InvalidCreateRepositoryException;

/**
 * Custom NPM Repository ( for private servers like verdaccio )
 *
 * @author Florian Bogner <f.bogner@pcsg.de>
 */
class CustomNpmRepository extends AbstractAssetsRepository
{
    /**
     * CustomNpmRepository constructor.
     * @param array $repoConfig
     * @param $io
     * @param $config
     * @param $eventDispatcher
     * @throws \Exception
     */
    public function __construct(array $repoConfig, $io, $config, $eventDispatcher)
    {
        if (!isset($repoConfig['custom_npm_url'])) {
            throw new \Exception('Missing parameter : custom_npm_url');
        }

        $this->url = $repoConfig['custom_npm_url'];

        parent::__construct($repoConfig, $io, $config, $eventDispatcher);
    }

    /**
     * {@inheritdoc}
     */
    protected function getType()
    {
        return 'npm';
    }

    /**
     * {@inheritdoc}
     */
    protected function getUrl()
    {
        return $this->url;
    }

    /**
     * {@inheritdoc}
     */
    protected function getPackageUrl()
    {
        return $this->canonicalizeUrl($this->baseUrl.'/%package%');
    }

    /**
     * {@inheritdoc}
     */
    protected function getSearchUrl()
    {
        return $this->canonicalizeUrl($this->baseUrl.'/-/all');
    }

    /**
     * {@inheritdoc}
     */
    public function search($query, $mode = 0, $type = null)
    {
        return [];
    }

    /**
     * {@inheritdoc}
     */
    protected function createVcsRepositoryConfig(array $data, $registryName = null)
    {
        $type = isset($data['repository']['type']) ? $data['repository']['type'] : 'vcs';

        return [
            'type'              => $this->assetType->getName().'-'.$type,
            'url'               => $this->getVcsRepositoryUrl($data, $registryName),
            'name'              => $registryName,
            'registry-versions' => isset($data['versions'])
                ? $this->createArrayRepositoryConfig($data['versions'])
                : [],
        ];
    }

    /**
     * {@inheritdoc}
     */
    protected function whatProvidesManageException(Pool $pool, $name, \Exception $exception)
    {
        if ($exception instanceof InvalidCreateRepositoryException) {
            $data = $exception->getData();

            if (isset($data['versions']) && !empty($data['versions'])) {
                $this->putArrayRepositoryConfig($data['versions'], $name, $pool);

                return;
            }
        }

        parent::whatProvidesManageException($pool, $name, $exception);
    }

    /**
     * Create and put the array repository with the asset configs.
     *
     * @param array $packageConfigs The configs of assets package versions
     * @param string $name The asset package name
     * @param Pool $pool The pool
     */
    protected function putArrayRepositoryConfig(array $packageConfigs, $name, Pool $pool)
    {
        $packages = $this->createArrayRepositoryConfig($packageConfigs);
        $repo     = new ArrayRepository($packages);
        Util::addRepositoryInstance($this->io, $this->rm, $this->repos, $name, $repo, $pool);

        $this->providers[$name] = [];
    }

    /**
     * Create the array repository with the asset configs.
     *
     * @param array $packageConfigs The configs of assets package versions
     *
     * @return CompletePackageInterface[]
     */
    protected function createArrayRepositoryConfig(array $packageConfigs)
    {
        $packages = [];
        $loader   = new ArrayLoader();

        foreach ($packageConfigs as $version => $config) {
            $config['version'] = $version;
            $config            = $this->assetType->getPackageConverter()->convert($config);

            try {
                $packages[] = $loader->load($config);
            } catch (\UnexpectedValueException $Exception) {
                if (isset($config['require'])) {
                    foreach ($config['require'] as $package => $versions) {
                        $config['require'][$package] = $this->cleanupVersions($config['require'][$package]);
                    }
                }

                if (isset($config['require-dev'])) {
                    foreach ($config['require-dev'] as $package => $versions) {
                        $config['require-dev'][$package] = $this->cleanupVersions($config['require-dev'][$package]);
                    }
                }

                try {
                    $packages[] = $loader->load($config);
                } catch (\UnexpectedValueException $Exception) {
                    echo '###################'.PHP_EOL;
                    echo $Exception->getMessage();
                    echo '###################'.PHP_EOL;
                    var_dump($config);
                    exit;
                }
            }
        }

        return $packages;
    }

    /**
     * @param $versionString
     * @return string|string[]
     */
    protected function cleanupVersions($versionString)
    {
        if (strpos($versionString, 'x') === false) {
            return $versionString;
        }

        $versions = explode(',', $versionString);

        foreach ($versions as $k => $version) {
            if (strpos($version, '>') === false
                && strpos($version, '<') === false
                && strpos($version, '=') === false
            ) {
                $versions[$k] = str_replace('.x', '.*', $version);
                continue;
            }

            $versions[$k] = str_replace('.x', '.0', $version);
        }

        $versionString = implode(',', $versions);

        return $versionString;
    }

    /**
     * Get the URL of VCS repository.
     *
     * @param array $data The repository config
     * @param string $registryName The package name in asset registry
     *
     * @return string
     *
     * @throws InvalidCreateRepositoryException When the repository.url parameter does not exist
     */
    protected function getVcsRepositoryUrl(array $data, $registryName = null)
    {
        if (!isset($data['repository']['url'])) {
            $msg = sprintf(
                'The "repository.url" parameter of "%s" %s asset package must be present for create a VCS Repository',
                $registryName,
                $this->assetType->getName()
            );
            $msg .= PHP_EOL.'If the config comes from the NPM Registry, override the config with a custom Asset VCS Repository';
            $ex  = new InvalidCreateRepositoryException($msg);
            $ex->setData($data);

            throw $ex;
        }

        return $this->convertUrl((string)$data['repository']['url']);
    }

    /**
     * Convert the url repository.
     *
     * @param string $url The url
     *
     * @return string The url converted
     */
    private function convertUrl($url)
    {
        if (0 === strpos($url, 'git+http')) {
            return substr($url, 4);
        }

        return $url;
    }
}
